from . import hash, tables
from .card import Card
from .evaluator import evaluate_cards
from .evaluator5 import evaluate_5cards
from .evaluator6 import evaluate_6cards
from .evaluator7 import evaluate_7cards
from .evaluator_omaha import evaluate_omaha_cards

__all__ = [
    hash,
    tables,
    Card,
    evaluate_cards,
    evaluate_5cards,
    evaluate_6cards,
    evaluate_7cards,
    evaluate_omaha_cards,
]



from __future__ import annotations

from typing import Union

# fmt: off
rank_map = {
    "2": 0, "3": 1, "4": 2, "5": 3, "6": 4, "7": 5, "8": 6, "9": 7,
    "T": 8, "J": 9, "Q": 10, "K": 11, "A": 12,
}
suit_map = {
    "C": 0, "D": 1, "H": 2, "S": 3,
    "c": 0, "d": 1, "h": 2, "s": 3
}
# fmt: on

rank_reverse_map = {value: key for key, value in rank_map.items()}
suit_reverse_map = {value: key for key, value in suit_map.items() if key.islower()}


class Card(int):
    def __new__(cls, value: Union[int, str, Card]) -> Card:
        if isinstance(value, str):
            rank, suit = value
            value = rank_map[rank] * 4 + suit_map[suit]
        return super(Card, cls).__new__(cls, value)

    def describe_rank(self) -> str:
        return rank_reverse_map[self // 4]

    def describe_suit(self) -> str:
        return suit_reverse_map[self % 4]

    def describe_card(self) -> str:
        return self.describe_rank() + self.describe_suit()

    def __str__(self) -> str:
        return self.describe_card()

    def __repr__(self) -> str:
        return f'Card("{self.describe_card()}")'



from typing import Iterable, Union

from .card import Card
from .evaluator5 import evaluate_5cards
from .evaluator6 import evaluate_6cards
from .evaluator7 import evaluate_7cards
from .evaluator_omaha import evaluate_omaha_cards


def evaluate_cards(*args: Iterable[Union[int, str, Card]]) -> int:
    cards = list(map(Card, args))

    if len(cards) == 5:
        return evaluate_5cards(*cards)

    elif len(cards) == 6:
        return evaluate_6cards(*cards)

    elif len(cards) == 7:
        return evaluate_7cards(*cards)

    elif len(cards) == 9:
        return evaluate_omaha_cards(*cards)



from .hash import hash_binary, hash_quinary
from .tables import FLUSH, FLUSH_OMAHA, NO_FLUSH_OMAHA

# fmt: off
binaries_by_id = [
    0x1, 0x1, 0x1, 0x1,
    0x2, 0x2, 0x2, 0x2,
    0x4, 0x4, 0x4, 0x4,
    0x8, 0x8, 0x8, 0x8,
    0x10, 0x10, 0x10, 0x10,
    0x20, 0x20, 0x20, 0x20,
    0x40, 0x40, 0x40, 0x40,
    0x80, 0x80, 0x80, 0x80,
    0x100, 0x100, 0x100, 0x100,
    0x200, 0x200, 0x200, 0x200,
    0x400, 0x400, 0x400, 0x400,
    0x800, 0x800, 0x800, 0x800,
    0x1000, 0x1000, 0x1000, 0x1000,
]
suitbit_by_id = [0x1, 0x8, 0x40, 0x200,] * 13
# fmt: on

# The first five parameters are the community cards
# The later four parameters are the player hole cards
def evaluate_omaha_cards(c1, c2, c3, c4, c5, h1, h2, h3, h4):
    value_flush = 10000
    value_noflush = 10000
    suit_count_board = [0, 0, 0, 0]
    suit_count_hole = [0, 0, 0, 0]

    suit_count_board[c1 & 0x3] += 1
    suit_count_board[c2 & 0x3] += 1
    suit_count_board[c3 & 0x3] += 1
    suit_count_board[c4 & 0x3] += 1
    suit_count_board[c5 & 0x3] += 1

    suit_count_hole[h1 & 0x3] += 1
    suit_count_hole[h2 & 0x3] += 1
    suit_count_hole[h3 & 0x3] += 1
    suit_count_hole[h4 & 0x3] += 1

    for i in range(4):
        if suit_count_board[i] >= 3 and suit_count_hole[i] >= 2:
            suit_binary_board = [0, 0, 0, 0]

            suit_binary_board[c1 & 0x3] |= binaries_by_id[c1]
            suit_binary_board[c2 & 0x3] |= binaries_by_id[c2]
            suit_binary_board[c3 & 0x3] |= binaries_by_id[c3]
            suit_binary_board[c4 & 0x3] |= binaries_by_id[c4]
            suit_binary_board[c5 & 0x3] |= binaries_by_id[c5]

            suit_binary_hole = [0, 0, 0, 0]
            suit_binary_hole[h1 & 0x3] |= binaries_by_id[h1]
            suit_binary_hole[h2 & 0x3] |= binaries_by_id[h2]
            suit_binary_hole[h3 & 0x3] |= binaries_by_id[h3]
            suit_binary_hole[h4 & 0x3] |= binaries_by_id[h4]

            if suit_count_board[i] == 3 and suit_count_hole[i] == 2:
                value_flush = FLUSH[suit_binary_board[i] | suit_binary_hole[i]]
            else:
                padding = [0x0000, 0x2000, 0x6000]

                suit_binary_board[i] |= padding[5 - suit_count_board[i]]
                suit_binary_hole[i] |= padding[4 - suit_count_hole[i]]

                board_hash = hash_binary(suit_binary_board[i], 5)
                hole_hash = hash_binary(suit_binary_hole[i], 4)

                value_flush = FLUSH_OMAHA[board_hash * 1365 + hole_hash]

            break

    quinary_board = [0] * 13
    quinary_hole = [0] * 13

    quinary_board[(c1 >> 2)] += 1
    quinary_board[(c2 >> 2)] += 1
    quinary_board[(c3 >> 2)] += 1
    quinary_board[(c4 >> 2)] += 1
    quinary_board[(c5 >> 2)] += 1

    quinary_hole[(h1 >> 2)] += 1
    quinary_hole[(h2 >> 2)] += 1
    quinary_hole[(h3 >> 2)] += 1
    quinary_hole[(h4 >> 2)] += 1

    board_hash = hash_quinary(quinary_board, 13, 5)
    hole_hash = hash_quinary(quinary_hole, 13, 4)

    value_noflush = NO_FLUSH_OMAHA[board_hash * 1820 + hole_hash]

    if value_flush < value_noflush:
        return value_flush
    else:
        return value_noflush



from struct import *
from pathlib import Path

DIR = Path(__file__).parent

FLUSH_OMAHA = unpack('<4099095h', open(DIR / 'omaha_flush.dat', 'rb').read())
NO_FLUSH_OMAHA = unpack('<11238500h', open(DIR / 'omaha_noflush.dat', 'rb').read())



from .dptables import CHOOSE, DP, SUITS
from .hashtable import FLUSH
from .hashtable5 import NO_FLUSH_5
from .hashtable6 import NO_FLUSH_6
from .hashtable7 import NO_FLUSH_7
from .hashtable_omaha import FLUSH_OMAHA, NO_FLUSH_OMAHA

__all__ = [
    CHOOSE,
    DP,
    SUITS,
    FLUSH,
    NO_FLUSH_5,
    NO_FLUSH_6,
    NO_FLUSH_7,
    FLUSH_OMAHA,
    NO_FLUSH_OMAHA,
]



from .hash import hash_quinary
from .tables import FLUSH, NO_FLUSH_5, SUITS

# fmt: off
binaries_by_id = [
    0x1, 0x1, 0x1, 0x1,
    0x2, 0x2, 0x2, 0x2,
    0x4, 0x4, 0x4, 0x4,
    0x8, 0x8, 0x8, 0x8,
    0x10, 0x10, 0x10, 0x10,
    0x20, 0x20, 0x20, 0x20,
    0x40, 0x40, 0x40, 0x40,
    0x80, 0x80, 0x80, 0x80,
    0x100, 0x100, 0x100, 0x100,
    0x200, 0x200, 0x200, 0x200,
    0x400, 0x400, 0x400, 0x400,
    0x800, 0x800, 0x800, 0x800,
    0x1000, 0x1000, 0x1000, 0x1000,
]
suitbit_by_id = [0x1, 0x8, 0x40, 0x200,] * 13
# fmt: on


def evaluate_5cards(a, b, c, d, e):
    suit_hash = 0

    suit_hash += suitbit_by_id[a]
    suit_hash += suitbit_by_id[b]
    suit_hash += suitbit_by_id[c]
    suit_hash += suitbit_by_id[d]
    suit_hash += suitbit_by_id[e]

    if SUITS[suit_hash]:
        suit_binary = [0] * 4

        suit_binary[a & 0x3] |= binaries_by_id[a]
        suit_binary[b & 0x3] |= binaries_by_id[b]
        suit_binary[c & 0x3] |= binaries_by_id[c]
        suit_binary[d & 0x3] |= binaries_by_id[d]
        suit_binary[e & 0x3] |= binaries_by_id[e]

        return FLUSH[suit_binary[SUITS[suit_hash] - 1]]

    quinary = [0] * 13

    quinary[a >> 2] += 1
    quinary[b >> 2] += 1
    quinary[c >> 2] += 1
    quinary[d >> 2] += 1
    quinary[e >> 2] += 1

    return NO_FLUSH_5[hash_quinary(quinary, 13, 5)]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['phevaluator', 'phevaluator.tables']

package_data = \
{'': ['*']}

install_requires = \
['numpy>=1.21.0,<2.0.0']

setup_kwargs = {
    'name': 'phevaluator',
    'version': '0.4.0.2',
    'description': 'Poker-Hand-Evaluator: An efficient poker hand evaluation algorithm and its implementation, supporting 7-card poker and Omaha poker evaluation',
    'long_description': None,
    'author': 'Henry Lee',
    'author_email': 'lee0906@hotmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



from .tables import CHOOSE, DP


def hash_quinary(q, length, k):
    sum_numb = 0

    for i in range(length):
        sum_numb += DP[q[i]][length - i - 1][k]

        k -= q[i]

        if k <= 0:
            break

    return sum_numb


def hash_binary(binary, k):
    sum_numb = 0
    length = 15

    for i in range(length):
        if binary & (1 << i):
            if length - i - 1 >= k:
                sum_numb += CHOOSE[length - i - 1][k]

            k -= 1

            if k == 0:
                break

    return sum_numb
